/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */

#ifdef _USRDLL
#define _LIBMIL_LIB_
#endif 

#if defined(_WIN32)
	#define DLLEXPORT __declspec(dllexport)
#else
	#define DLLEXPORT extern
#endif

#if !defined(_LIBMIL_LIB_)
	#define MIL_API DLLIMPORT
#else
	#define MIL_API DLLEXPORT
#endif

// disable error related to dll export of STL method 
#pragma warning (disable: 4251)

#include <Python.h>
#include <FLOAT.h>
#include <map>
#include "matrix.h"
#include <stdlib.h>
#include <math.h>
#include <malloc.h>
#include <algorithm>
#include <vector>
#include <strstream>
#include <sstream>
#include <iostream>
#include <hash_map>
#include <list>
#include <fstream>
#include <set>
#include <iomanip>

#include "qublib.h"

#include <QUB_Tree.h>
#include <QTR_NodeFunc.h>
#include <qubtree_node.h>
#include <qubtree_NodeFunc.h>

#include "qub_reportstream.h"

#include "milutil.h"
#include "matrixutil.h"

using stdext::hash_map;

//----- Model Structures   (Formerly ch_x.h)
	typedef struct tag_mdlinf{
		fq::vector<int> clazz; // class of state i		
		fq::matrix<int> path;    // [from, to, drug] of path i		
		fq::vector<double> x ; // [k0, k1] for each path		
		fq::vector<double> pr;  // start prob of state i		
		
		int nstate;  // derived		
		int npath;   // derived // one-directional		
		
		fq::vector<double> xlimit[2];		
		fq::matrix<double> mtx;		
		fq::vector<double> vct;		
		} mdlinf;

//----- From dfp_optimize
	class datinf {
		public:
			int nchannel;		
			fq::vector<double> i;       // single amplitudes		
			fq::matrix<double> r;       // single correlations		
		};

	class dfp_optimizable{
		public:
			virtual ~dfp_optimizable() {}
			
			virtual bool getStartingZ( fq::vector<double> &z ) = 0;
			virtual bool evaluate( fq::vector<double> &z, fq::vector<double> &gz,
							double &ll, int &nf, int &ndf, bool flagGrad=true ) = 0;
			virtual bool checkContinue( int iter, int nf, int ndf, double ll,
							fq::vector<double> &z, fq::vector<double> &gz )=0;
		};

	class dfp_result{
		public:
			fq::vector<double> z;
			double ll;
			int iter;
			int nfcn, ndfcn;
			fq::matrix<double> hessian;
			int err; // ? 
		};

	dfp_result dfp_optimize( dfp_optimizable *optable, int maxIter, double convLL, 
							double convGrad, double stepMax = 1.0 );

	// dll exported for delphi.	Both return 0 for success
	typedef int (__stdcall *dfp_dll_func)(void*, double*, double*, double*, int*, int*, int);
	//                          caller, params,  grads,   ll,      nf,   ndf,  (bool) computeGrads
	typedef int (__stdcall *dfp_dll_check)(void*, int, int, int, double, double*, double*);
	//                          caller, iters, nf, ndf,   ll     params,  grads



	extern "C" MIL_API
	double __stdcall dfp_opt_dll( void *call_obj, int npar, double *initPar, double **hessianOut,
						dfp_dll_func func, dfp_dll_check check,
						int maxIter, double convLL, double convGrad, double stepMax,
						int *errOut );
		// errOut:  0   success
		//          -1  func returned error
		//          -2  check returned error
		//          -3  exceeded max iter

	//----- From ModelTree
	class ModelTree	{
		public:
			ModelTree();
			ModelTree( QUB_Tree mdlNode );
			ModelTree( const ModelTree& copy );
			virtual ~ModelTree();

			QUB_Tree node;
			hash_map<string, int> indexOfParam;
			int nparam;
			std::vector<int> cns_src_ix;

			void updateParams(); // if a rate's ligand or voltage could have changed since construction
			void verifyArs(); // something in qub likes to remove unused classes but a lot of old code here depends on 'em

			int toMdlinf( mdlinf& mi );
			// just nchannel, amp, sd, ar
			void toDatinf( datinf& di );
			void fromDatinf( datinf& di ); // same nclass, nar (meant for amp updates)

			// old algos don't work if there are no states of class 0 (or any class < max used class)
			// do this before .toMdlinf(), restore after committing changes
			void condenseClasses();
			void restoreClasses();
			int  classesPresent();

			void updateRates( mdlinf& mi );
			void updateRates( mdlinf& mi, fq::vector<double>& xsd );

			// update amp, sd, ar

			// generate metamodel?
		};

	extern "C" MIL_API QTR_Impl* __stdcall DoMergeModels( QTR_Impl *factors, QTR_INodeCFunc *reportCB );
	extern "C" MIL_API int __stdcall DoUnmergeModels( QTR_Impl *product, QTR_Impl *factors, QTR_INodeCFunc *reportCB );

	extern "C" PyObject * DoMergeModelsPy( PyObject *self, PyObject *args );
	extern "C" PyObject * DoUnmergeModelsPy( PyObject *self, PyObject *args );

	
	extern "C" MIL_API QTR_Impl* __stdcall BalanceAllLoops( QTR_Impl *model );
	extern "C" PyObject * DoBalanceAllLoopsPy( PyObject *self, PyObject *args );
	QUB_Tree DoBalanceAllLoops(QUB_Tree model);

	extern "C" MIL_API QTR_Impl* __stdcall PareConstraints( QTR_Impl *model, QTR_Impl *removed );
	extern "C" MIL_API QTR_Impl* __stdcall PareLoopConstraints( QTR_Impl *model );
	extern "C" PyObject * DoPareConstraintsPy( PyObject *self, PyObject *args );
	QUB_Tree DoPareConstraints(QUB_Tree model, QUB_Tree removed);
	QUB_Tree DoPareLoopConstraints(QUB_Tree model);

	extern "C" MIL_API QTR_Impl* __stdcall CheckLoopStimuli( QTR_Impl *model );
	extern "C" PyObject * DoCheckLoopStimuliPy( PyObject *self, PyObject *args );
	QUB_Tree DoCheckLoopStimuli(QUB_Tree model);

	QUB_Tree GetScaledRates(QUB_Tree constraints, int from, int to);
	QUB_Tree GetScaledExpRates(QUB_Tree constraints, int from, int to);
	extern "C" MIL_API QTR_Impl* __stdcall _GetScaledRates(QTR_Impl* constraints, int from, int to);
	extern "C" MIL_API QTR_Impl* __stdcall _GetScaledExpRates(QTR_Impl* constraints, int from, int to);


//-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
//----- From HistPdf.cpp
	void binning(float ts, int iSegs, int* ndwell, float** tdwell, int* nbin, float* bin, float tdead, double& rOut);
	void calhist(float ts, int ic, int iSegs, int* ndwell, int **idwell, float** tdwell, int iBins, float* bin, float* hst, float tdead=0.0f);
	void calhist_smooth(int ic, int iSegs, int* ndwell, int **idwell, float** tdwell, int iBins, float* bin, float* hst, float tdead, float sampling, float r);

//----- From QMatrixUtil
	class CChannelCombine { 
		private :
			int m_iChannels;
			int m_iStates;
			int m_iTemp;
			int m_iStateIndex;
			bool m_lContinue;
			int * m_aStates;
		public : 
			// Initialize and create the first combination
			CChannelCombine(int iChannels, int iStates);
			// Check if another combination exists 
			bool GetNextCombine();
			// Access the state elements with a 0 based index ( internally stored as 1 based array ) 
			int State0(int i);
			~CChannelCombine();
		};

	void CalcMultiPr(int nchannel, int nstate, int nmetastate, int **metastate, double *pr, double* mprx);
	void CalcMultiPr_mpl(int nchannel, int nstate, int nmutistate, int **mutistate, double *pr, double* mprx);
	double lnfactrl(int n);
	void metaMarkov(int nchannel, int nstate, int npath, int **path, int *class1, 
					float *ratio, int *nmetastate, int *nmetapath, int *nmetaclass, 
					float *metaamp, int **metastate, int **metapath, int *metaclass, 
					int **subpath);
	BOOL freePar(int mc, int nx, double *x, double **Mtx, double *vct, int* nz, double *z, ostream &msgOut);
	BOOL mil_ztox(int nz, int nx, double **Mtx, double *vct, double *z, double *x, double **dx_z, double *xmin, double *xmax);
	void qtometaq(int nstate, int npath, int **path, int nmetastate,
				  int nmetapath, int **metapath, int **metastate, int **subpath,
				  double **q, double **metaq, double **dmetaq_q);
	void aggregate(int nstate, int nclass, int *class1, int *ngroup, int **index);
	BOOL qspctrm(int n, double **q, double* wr, double **s, double **s1, int &posieigencount, ostream &msgOut);
	void qtoqe(int nstate, int npath, int **path, int *class1, int nclass, double td, double **q, double **qe, double ***dqe_q);
	void mexp(int n, float t, float* q, float* a, int nz, float* dq, float* da, int i0) ;
	float gaussian(float mean, float xms, float x);
	void qtomutiq(int nstate, int npath, int* path, int nmutistate, 
				  int* mutistate, int nmutipath, int* mutipath, float* q, 
				  float* mutiq, int nz, float* dq, float* dmutiq);
	void xtoq(int nstate, int npath, int* path, float* drug, float volt, 
			  float* x, float* q, int nz, float* dx, float* dq);
	void eclass(float * aData, int iData, int & iClasses, float * aClasses, int * ia, double openDirection);


//----- From MSearchEvalTree.cpp
	extern "C" MIL_API QTR_Impl * __stdcall msearch_tree_iface( QTR_Impl *cfg, 
				QTR_Impl **data, QTR_Impl *models, QTR_INodeCFunc *rptCB, 
				QTR_INodeCFunc *rsltCB, QTR_INodeCFunc *pctCB, QTR_INodeCFunc *filterCB );

	extern "C" PyObject * msearch_tree_Pyiface( PyObject *self, PyObject *args );

	extern "C" MIL_API void __stdcall msearch_modelCount_iface( QTR_Impl *model, 
				QTR_Impl *models, int *allModelCount, int *noLoopCount );


//----- From Mil_Eval_Tree.cpp
	class mil_model{
		public:
			mil_model( QUB_Tree modelNode, double searchLimit ); // changes a copy of modelNode

			void reset(); // re-copy original

			void updateRates(); // from mi.x into node
			void updateRates( fq::vector<double>& xsd );
			
			QUB_Tree  original;
			QUB_Tree  node; // the one used by tree
			ModelTree tree;
			double    searchLimit;
			mdlinf    mi;
			int       nconstraint;
			int       nclass;
			int       nchannel;

		private:
			void normalizePr();
		};

	class mil_metamodel{
		public:
			mil_metamodel( mil_model& mdl, int nchan );

			mil_model&       model;
			int              nchannel;

			// generated on construction:
			fq::vector<int>    idrug, ivolt; // idrug[i] = mi.path[i][2], ivolt[i] = mi.path[i][3]

			int              nmetastate, nmetaclass, nmetapath;
			fq::vector<int>    metaclass, nmetagroup;
			fq::vector<float>  metaamp;
			fq::vector<double> mprx;
			fq::matrix<int>      metastate, metapath, subpath, metaindex;
			int              maxnmetastate, maxnmetapath;
		};

	void UpdateResult( QUB_Tree result, double ll, fq::vector<double>& gz );
	void BlessResult( QUB_Tree result );

	class mil_workspace { 
			// can evaluate segments and make histograms for one file (dataset with same conditions)
		public:
			mil_workspace( QUB_Tree dataset, mil_model& mdl, bool usePequil, bool smoothHistoBins, bool sqrtHisto, int *StopFlag, std::ostream& errstream );

			void reset(); // back to original rates

			int segCount();
			int classCount(); // in metamodel
			int classCountInData();
			int componentsInClass( int cls );

			// evaluate(...) is where the action is.
			// evaluate multi workspaces = sum( eachWorkspace.evaluate() )
			// call updateZ() first if z has changed
			// returns number of failed segments
			int evaluate( fq::vector<double>& gz, double& ll, int flagGrad, int seg=-1 ); // default: all segs

			// provide the nodes that hold the "Histogram" and "TimeConstants" elements.
			// you can use them repeatedly.
			void getHistograms( QUB_Tree histGroup, int seg=-1 ); // bins n' bars
			// you provide # of bins in histGroup["BinCount"]
			void getPDFs( QUB_Tree histGroup, QUB_Tree timeConstContainer ); // pdf and TCs for given bins

			QUB_Tree getModelStats(double tMax, int tBins);

			bool resetZ( bool alsoResetModel=true ); // if false, you get the modified rates
			bool updateZ( fq::vector<double>& newz );
			void updateModel(); // update tree from most recent z
			void updateSD( fq::matrix<double>& hessian ); // into the tree

			fq::vector<double> z;      // constrained parameters
			bool             zValid; // false if the initial guess didn't meet constraints

			QUB_Tree         data; // with dead time applied and subtracted
			std::vector<QUB_Tree> segments;

		protected:
			void copyMetaData( QUB_Tree orig );
			void copyData( QUB_Tree orig );
			void copyDataJoined( QUB_Tree orig );

			void setupArrays();

			void setTimeConstHint( QUB_Tree hist, QUB_Tree tc );

			void calHist( int& nbin, float *bins, float *hist, int cls, int seg=-1 ); // can decrease nbin
			bool calPDFandTC( int nbin, float *bins, float **pdf, double *tau, double *amp, double& meanTau, int cls );

			void peqm(double **q, int ic, double *pe);
			bool pdfcof(double **q, int ic, double *pe, double *c, double *wr);
			void pdfbin(int ic, double *w, double *c, double td, int nbin, float* bin, float **pdf);

			void getModelStats(double tMax, int tBins, double **q, string suffix, QUB_Tree result);

			void setupDQDQ();          // path            -> dQdq
			void calcStartProbs();     // q, dQdq         -> pr, dpr_q
			void calcMultiProbs();     // pr, metastate   -> mpr, dmpr_pr
			void calc_dlogLL_q( int cls, fq::vector<double>& dLL_q );

			int              maxndwt;
			std::vector<int>      ndwt;
			std::vector<int*>     idwt;
			std::vector<float*>   tdwt;
			int              classesInData;

			mil_metamodel    model;
			bool             usePeq; // for start probs

			int              *Stopped;
			std::ostream&    milerr;

			fq::vector<double> C; // , V;  C now has concs and voltages intermingled
			double           td;

			// constraints: x = mtx * z + vct
			fq::matrix<double>   mtx;
			fq::vector<double> vct;

			int              posieigencount;
			fq::vector<double> df_q, df_x, df_z, df_qq;
			fq::matrix<double>   q, qq, qqe, dx_z, dq_x, dqq_q, df_qqe;
			fq::tensor<double>   dqqe_qq;
			fq::matrix<double*>  dqqe_qqAsPtr;
			fq::matrix<double>   dmpr_pr, dpr_q, dmpr_q;
			fq::vector<double> v;
			fq::matrix<double>   w, a, c, c1;
			fq::tensor<double>   p, p1;
			fq::tensor<double>   dQdq; // as in MIP, for dpr_q

			fq::matrix<double>   alpha, beta;
			fq::vector<double> scale;

			bool             smoothHistBins, sqrtHist;
			double           histBinR;
		};

	typedef CountedPtr<mil_workspace> mil_worksptr;


	class mil_eval_tree : public dfp_optimizable{
		public:
			mil_eval_tree( std::vector<QUB_Tree>& dataSets, QUB_Tree config,
							QTR_NodeFptr rptCB, QTR_NodeFptr rsltCB, QTR_NodeFptr pctCB );

			QUB_Tree execute();

			virtual bool getStartingZ( fq::vector<double> &z );
			virtual bool evaluate( fq::vector<double> &z, fq::vector<double> &gz,
									double &ll, int &nf, int &ndf, bool flagGrad=true );
			virtual bool checkContinue( int iter, int nf, int ndf, double ll,
										fq::vector<double> &z, fq::vector<double> &gz );

		protected:
			void initResults();
			void addTCSelectVars( QUB_Tree seg );
			void addRateSelectVars( QUB_Tree seg );
			void finishResults();

			void updatePDFs();
			void sendResult( QUB_Tree resNode, int iter, bool final=false, int errCode = 0 );

			bool canRun();
			bool evaluateAll( fq::vector<double> &z, fq::vector<double> &gz,
							  double &ll, int &nf, int &ndf, bool flagGrad=true );

			dfp_result runDFP();

			mil_model model;
			std::vector<mil_worksptr> workspaces;
			QUB_Tree cfg;
			QTR_NodeFptr reportCB, resultCB, percentCB;

			mil_reportstream milerr;

			bool opt, together, pdfsEachIter;
			int maxIter, maxRestarts;
			double maxStep, convLL, convGrad;
			
			int *Stopped;

			mil_worksptr currSpace; // for together == false
			int currSeg, iSeg, nSeg;
			int iterations;
			
			QUB_Tree result;
			QUB_Tree pctNode;
			int *pctPtr;
		};

	extern "C" MIL_API QTR_Impl * __stdcall miltreeiface( QTR_Impl *cfg, QTR_Impl **data,
				  QTR_INodeCFunc *rptCB, QTR_INodeCFunc *rsltCB, QTR_INodeCFunc *pctCB );

	extern "C" PyObject * miltreePyiface( PyObject *self, PyObject *args );

	typedef struct {
		int              nclass;
		int              *classOfState; // aka mi->clazz

		int              nmetastate;
		int              nmetaclass;
		int              nmetapath;
		int              maxnmetastate;
		int              maxnmetapath;

		int              *metaclass;
		int              *nmetagroup;
		double           *metapr;
		int              **metastate;
		int              **metapath;
		int              **subpath;
		int              **metaindex;

		mil_model        *model;
		mil_metamodel    *metamodel;
		} mil_export_model;

	extern "C" MIL_API mil_export_model * __stdcall new_mil_model( QTR_Impl *model, int nchannel );
	extern "C" MIL_API void __stdcall del_mil_model( mil_export_model *mdl );

//----- From MDLREP.CPP
	void mil_expQ_full(int n, double* w, double** p, double** p1, double t, double** a);


//----- From MPL_eval_tree
	class mpl_metamodel{
		public:
			mpl_metamodel( mil_model& mdl, int nchan );
			~mpl_metamodel();

			mil_model&       model;
			int              nchannel;

			// generated on construction:
			fq::vector<int>    idrug; // idrug[i] = mi.path[i][2]
			fq::matrix<double>   xlimit; // mi.xlimit transposed
			fq::vector<double> ratio; // ratio[i] == i

			fq::vector<double> amp, xms;
			fq::matrix<double>   ar;
			int              nar;

			int              nmutistate, nmuticlass, nmutipath;
			int              **mutistate, **mutipath;
			fq::vector<int>    muticlazz;
			fq::vector<double> mutipr;
		};

	extern "C" MIL_API QTR_Impl * __stdcall
	mpltreeiface( QTR_Impl *cfg, QTR_Impl **data, QTR_INodeCFunc *fillDataCB,
				  QTR_INodeCFunc *rptCB, QTR_INodeCFunc *rsltCB, QTR_INodeCFunc *pctCB );

	extern "C" PyObject * mpltreePyiface( PyObject *self, PyObject *args );
	// it will work with a bogus fillDataCB as long as the data is already there

//----- From Amp_eval_tree
	extern "C" MIL_API QTR_Impl * __stdcall amptreeiface( QTR_Impl *cfg, QTR_Impl **data,
				  QTR_INodeCFunc *fillDataCB, QTR_INodeCFunc *rptCB, QTR_INodeCFunc *pctCB );

	extern "C" PyObject * amptreePyiface( PyObject *self, PyObject *args );
					// it will work with a bogus fillDataCB as long as the data is already there


//----- From Stat_eval_tree
	QUB_Tree skm_ultd_runStat( QUB_Tree dataSet, QUB_Tree config, QTR_NodeFptr fillDataCB, QTR_NodeFptr reportCB );
	extern "C" MIL_API QTR_Impl * __stdcall stattreeiface( QTR_Impl *cfg, QTR_Impl **data, QTR_INodeCFunc *fillDataCB, QTR_INodeCFunc *rptCB );
	extern "C" PyObject * stattreePyiface( PyObject *self, PyObject *args );
	QUB_Tree ApplyDeadTime(QUB_Tree DataSet, float tdMS);

//----- From SKM Eval Tree 
	class skm_model{
		public:
			skm_model( QUB_Tree modelNode, int *fFixAmp=0, int *fFixSD=0 );
			
			QUB_Tree        node;
			ModelTree       tree;
			mdlinf          mi;

			int             nclass, nchannel;
			fq::vector<int>   path, fixAmp, fixSD;
			fq::vector<float> x, q, amp, xms;
		};

	class skm_metamodel{
		public:
			skm_metamodel( skm_model& mdl, int nchan, double dtMS );
			~skm_metamodel();

			skm_model& model;
			int nchannel;

			int             nmutistate, nmuticlass, nmutipath;
			int             *mutistate, *mutipath; // allocated in skm_mutimakv() [skm-utils.cpp]
			fq::vector<int>   cmutistate, mutiFixAmp, mutiFixSD;
			fq::vector<float> mutiq, mutia, mutipr, mutiamp, mutixms;
		};

	QUB_Tree skm_unlimited( std::vector<QUB_Tree>& dataSets, QUB_Tree config,
							QTR_NodeFptr fillDataCB, QTR_NodeFptr reportCB, QTR_NodeFptr pctCB, bool doStat = false );

	extern "C" MIL_API QTR_Impl * __stdcall
		skmtreeiface( QTR_Impl *cfg, QTR_Impl **data,
				  QTR_INodeCFunc *fillDataCB, QTR_INodeCFunc *rptCB, QTR_INodeCFunc *pctCB );

	extern "C" PyObject * skmtreePyiface( PyObject *self, PyObject *args );
	// it will work with a bogus fillDataCB as long as the data is already there

//----- 6/2005 - Expression Evaluation engine 
	/*
	The following two functions provide a simple interface to a global CExpression 
	object which can be used for c style dll access.
	*/
	extern "C" DLLEXPORT void __stdcall LoadExpression( const char * pchExpr );
	extern "C" DLLEXPORT double __stdcall EvaluateExpression( double t );

	/* 
	Expression parser / evaluator.
	Evaluate f(t) for provided t.  
	f = Expression of + - * / ^ ln exp sin cos ( ) abs sgn
	'-' preceded by operator or '(' is negate, otherwise it is minus
	Ex : 
		CExpression o;
		o.Parse("t^3-60*t^2+1100*t-6000+abs(sin(t/5)*sin(t/4)*100+90*cos(t/6)*cos(t/3))");
		double t1=o.Evaluate(0.532);
		double t2=o.Evaluate(9876.5432);
	*/
	class CExpression { 
		private : 
			int m_iTokens, m_iStackSize;
			int * m_aTokens;		// Expression parsed and reordered w/o parentheses
			int * m_aRaw;			// Expression parsed in original order 
			double * m_aValues;
			double * m_aStack;
			bool IsFunction( int iToken );
			bool IsValue( int iToken );
			bool IsBinary( int iToken );
			int ParseFind( int iToken1, int iToken2, int i0, int i1 ); 
			int ParseFindL( int iToken1, int iToken2, int i0, int i1 );
			void ParseShiftL( int i0, int i1, int iShift=1 );
			void ParseReduce( int i0, int i1 );
			bool NextToken( const char * pchExpr, int * piPos, int * piToken, int iPrevToken, double * pnValue );

		public :
			CExpression( const char * pchExpr );
			CExpression();
			~CExpression();
				
			bool Parse( const char * pchExpr ); 
			double Evaluate( double t );

			#ifdef _DEBUG
			//----- This function displays the parse results for debug purposes
			void Text( char * achBuffer, int iBufSize );
			#endif
		};
